'''
Created with love by Sigmoid.
@Author - Spînu Lavinia - laviniaspinu01@gmail.com
'''
# Import all needed libraries.
import pandas as pd
from sklearn.decomposition import FastICA
from .errors import MissingDataError, NonNumericDataError, NoSuchColumnError, DifferentColumnsError

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

class ICAFilter():
    def __init__(self, n_components : int = None):
        '''
            Setting the algorithm
        :param n_components: integer, by default = None
            Number of components to keep
        '''
        self.n_components = n_components

    def filter(self, dataframe : pd.DataFrame, target : str):
        '''
            Creating filter to new data and separating superimposed signals
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: string
             The column name of the value that we have to predict
        '''
        # Checking for missing values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values."
            )

        # if a column has string values raise an error
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # if the given target doesn't match any column in the DataFrame raise an error
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist."
            )

        # Splitting the dataframe.
        self.dataframe = dataframe.copy()
        self.target = target
        self.X_columns = [col for col in self.dataframe.columns if col != self.target]
        self.X = self.dataframe[self.X_columns].values
        self.y = self.dataframe[target].values

        # Creating the filter.
        self.ica = FastICA(n_components = self.n_components)
        self.ica.fit(self.X)

        # Creating new data based on the filter.
        X_ica = self.ica.transform(self.X)
        X_new = self.ica.inverse_transform(X_ica)
        X_new = pd.DataFrame(X_new, columns=self.X_columns)

        # Create and return new Dataframe.
        X_new[target] = self.y
        return X_new

    def apply(self, dataframe : pd.DataFrame):
        '''
            Separating superimposed signals
            based on an already existed filter
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: string
             The column name of the value that we have to predict
        '''
        # Checking for missing values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values."
            )

        # if a column has string values raise an error
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Splitting the data.
        self.data_frame = dataframe.copy()

        # Checking if the names of the columns are different for the both dataframes.
        if list(self.data_frame.columns) != list(self.dataframe.columns):
            raise DifferentColumnsError('The passed dataframe has different columns from the one passed to filter function.')
        X_columns = [col for col in self.data_frame.columns if col != self.target]
        X = self.data_frame[X_columns].values
        y = self.data_frame[self.target].values

        # Applying filter to the new dataframe.
        X_ica = self.ica.transform(X)
        X_new = self.ica.inverse_transform(X_ica)
        X_new = pd.DataFrame(X_new, columns=self.X_columns)

        # Create and return new Dataframe.
        X_new[self.target] = y
        return X_new



'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.feature_selection import chi2
from .errors import MissingDataError, NoSuchColumnError, NonNumericDataError

class ChiSquaredSelector:
    def __init__(self, significance_level : float = 0.05) -> None:
        '''
            Setting up the algorithm
        :param significance_level: float, default = 0.05
            Used to compare the p-values with to find the features that must be removed from model
        '''
        self.significance_level = significance_level
        self.iter = 0

    def bin_to_cols(self, feature_state : list, X_column : list) -> list:
        '''
            Converting the binary state list into names of features that are included in model
        :param feature_state: list
            A list with zeros and ones used by model to understand what values should be pick by the model
            (when value = 1) and what it shouldn't pick (when value = 0)
        :param X_column: list
            The list with features of the dataframe before the algorithm is applied
        :return: list
            The list of features that where picked by the algorithm
        '''
        included_columns = [X_column[i] for i in range(len(feature_state)) if feature_state[i]==1]
        return included_columns

    def select(self, dataframe : 'pd.DataFrame', target : str) -> list:
        '''
            Selecting the most important columns
        :param dataframe: pandas DataFrame
            Data Frame on which the algorithm is applied
        :param target: string
            The column name of the value that we what to predict
        :return: list
            The list of features that are selected by the algorithm as the best one
        '''
        # Checking if in the passed data are any NaN values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values!"
            )

        # Checking if in the passed data are non-numeric values.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Checking if the passed target column name is present in the passed data frame.
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist!"
            )

        self.dataframe = dataframe.copy()

        # Getting the list with names of columns without the target one
        X_columns = [col for col in self.dataframe.columns if col != target]

        # Creating the ch2 and p-value history dictionaries
        self.chi2_history = {}
        self.p_value_history = {}
        for col in X_columns:
            self.chi2_history[col] = []
            self.p_value_history[col] = []

        # Defining the feature states
        feature_state = list(np.ones(len(X_columns)))
        while True:
            self.iter +=1

            # Extracting the selected columns
            X_cols = self.bin_to_cols(feature_state, X_columns)
            self.X = self.dataframe[X_cols].values
            self.y = self.dataframe[target].values

            # Getting the chi2 and p-values
            chi2_vals, p_vals = chi2(self.X, self.y)
            index = 0
            for col in X_columns:
                if col in X_cols:
                    self.chi2_history[col].append(float(chi2_vals[index]))
                    self.p_value_history[col].append(float(p_vals[index]))
                    index+=1
                else:
                    self.chi2_history[col].append(-1)
                    self.p_value_history[col].append(-1)

            # Erasing the column with the p-value equal with the max value of the p-value, if the max value is
            # higher than significance level
            max_PValue = max(p_vals)
            if max_PValue > self.significance_level:
                for j in range(len(X_cols)):
                    if p_vals[j].astype(float) == max_PValue:
                        feature_state[X_columns.index(X_cols[j])] = 0
            else:
                break
        self.choosed_cols = self.bin_to_cols(feature_state, X_columns)

        # Returning the chose columns.
        return self.choosed_cols

    def plot_chi2(self, title : str = "Chi2 Plot", save : bool = False, file_path : str = None) -> None:
        '''
            Plotting the process of selecting the important features using chi2.
        :param title
            The title of the plot
        :param save: boolean, default = False
            If the this parameter is set to False that the model will not save the model
            If it is set to True the plot will be saved using :param file_path
        :param file_path: string, default = None
            The file path where the plot will be saved
            If the :param save is set to False the it is not used
        '''
        for col in self.chi2_history:
            line_style = lambda col: '-' if col in self.choosed_cols else '--'
            plt.plot(range(self.iter), self.chi2_history[col], line_style(col), label=col)
        plt.xlabel("Iterations")
        plt.ylabel("Chi2 Score")
        plt.legend(ncol=3)
        plt.title(title)
        if save:
            plt.savefig(file_path)
        plt.show()

    def plot_p_value(self, title : str = "P-Value Plot", save : bool = False, file_path : str = None) -> None:
        '''
            Plotting the process of selecting the important features using p-values.
        :param save: boolean, default = False
            If the this parameter is set to False that the model will not save the model
            If it is set to True the plot will be saved using :param file_path
        :param file_path: string, default = None
            The file path where the plot will be saved
            If the :param save is set to False the it is not used
        '''
        line_style = lambda col: '-' if col in self.choosed_cols else '--'
        for col in self.chi2_history:
            plt.plot(range(self.iter), self.p_value_history[col], line_style(col), label=col)
        plt.xlabel("Iterations")
        plt.ylabel("p-values")
        plt.legend(ncol=3)
        plt.title(title)
        if save:
            plt.savefig(file_path)
        plt.show()


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
import pandas as pd
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from .errors import NoSuchMethodError, MissingDataError, NoSuchColumnError, NonNumericDataError, NotBetweenZeroAndOneError, DifferentColumnsError

class LDAReducer:
    def __init__(self, solver : str = 'svd', method : str = 'pearson', min_corr : float = 0.5, max_corr : float = 0.8) -> None:
        '''
            The constructor of the reducer.
        :param solver: str, default = 'svd'
            The solver used by LDA Algorithm.
        :param method: str, default = 'person'
            The type of correlation used to find the most correlated features.
        :param min_corr: float, default = 0.5
            The minimal correlation between features accepted.
        :param max_corr: float, default = 0.8
            The maximal correlation between features accepted.
        '''
        # Setting up the Reducer
        if solver in ['svd', 'lsqr', 'eigen']:
            self.solver = solver
        else:
            raise NoSuchMethodError(f"kydavra doesn't sustain such solver as {solver}/nTry 'svd', 'lsqr' or 'eigen'")
        if min_corr < 0 or max_corr > 1 or min_corr > 1 or max_corr < 0:
            raise NotBetweenZeroAndOneError("Min or Max Correlations are not seted between 0 and 1!")
        else:
            self.min_corr = min_corr
            self.max_corr = max_corr
        if method in ['pearson', 'kendall', 'spearman']:
            self.method = method
        else:
            raise NoSuchMethodError(f"kydavra doesn't sustain such method as {method}/nTry 'pearson', 'kendall' or 'spearman'")

    def reduce(self, dataframe : pd.DataFrame, target : str) -> pd.DataFrame:
        '''
            This function reduces the correlated columns using LDA.
        :param dataframe: pandas DataFrame
            The pandas DataFrame on which the function is applied.
        :param target: str
            The target column name.
        :return: pandas DataFrame
            The data frame with reduced correlated columns.
        '''
        # Checking if in the passed data are any NaN values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values!"
            )

        # Checking if in the passed data are non-numeric values.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Checking if the passed target column name is present in the passed data frame.
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist!"
            )

        self.df = dataframe.copy()
        # Extracting X column names
        self.X_columns = [col for col in self.df.columns if col != target]

        # Calculating the correlation table
        self.corr_table = self.df[self.X_columns].corr(method=self.method).values

        # Searching for all correlated pairs.
        self.correlated_pairs = []
        for i in range(len(self.corr_table)):
            for j in range(i):
                if abs(self.corr_table[i][j]) < self.max_corr and abs(self.corr_table[i][j]) > self.min_corr:
                    self.correlated_pairs.append(tuple([self.X_columns[i], self.X_columns[j]]))

        # BUilding a set with unique correlated columns
        self.correlated_cols = set()
        for pair in self.correlated_pairs:
            self.correlated_cols.add(pair[0])
            self.correlated_cols.add(pair[1])

        # Grouping correlated columns into groups
        self.correlated_groups = dict()
        for col in list(self.correlated_cols):
            self.correlated_groups[col] = []
            for pair in self.correlated_pairs:
                if col in pair:
                    self.correlated_groups[col].append(pair[0] if pair[0] != col else pair[1])
        self.correlated_groups = [tuple([key] + self.correlated_groups[key]) for key in self.correlated_groups]
        self.correlated_groups = [tuple(sorted(group)) for group in self.correlated_groups]
        self.correlated_groups = list(set(self.correlated_groups))

        # Sorting groups to start
        self.correlated_groups.sort(key=len, reverse=True)
        self.ldas = dict()
        for group in self.correlated_groups:
            # Checking if columns are present in data frame.
            present = True
            for col in group:
                if col in self.df.columns:
                    pass
                else:
                    present = False
            if not present:
                break
            # Building the model in column group
            lda = LinearDiscriminantAnalysis(solver=self.solver)
            X = self.df[list(group)].values
            y = self.df[target]
            lda.fit(X, y)
            new_column_name = '_'.join(group)
            self.df[new_column_name] = lda.predict(X)
            self.df = self.df.drop(list(group), axis=1)
            self.ldas[group] = lda
        return self.df

    def apply(self, dataframe : pd.DataFrame) -> pd.DataFrame:
        '''
            This functions applies the trained LDAs on the new data set.
        :param dataframe: pandas DataFrame
            The pandas DataFrame on which the function is applied.
        :return: pandas DataFrame
            The data frame with reduced correlated columns.
        '''
        # Checking if in the passed data are any NaN values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values!"
            )

        # Checking if in the passed data are non-numeric values.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Checking if the names of the columns are different for the both dataframes.
        if list(self.df.columns) != list(dataframe.columns):
            raise DifferentColumnsError('The passed dataframe has different columns from the one passed to filter function.')

        # Applying the LDAs on the new data frame.
        self.df = dataframe.copy()
        for group in self.ldas:
            X = self.df[list(group)].values
            self.df['_'.join(group)] = self.ldas[group].predict(X)
            self.df = self.df.drop(list(group), axis=1)
        return self.df


'''
Created with love by Sigmoid

@Author - Vladimir Stojoc - vladimir.stojoc@gmail.com
@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
import numpy as np
import matplotlib.pyplot as plt
from sklearn.linear_model import ElasticNetCV, enet_path
import pandas as pd
from .errors import NoSuchColumnError, NonNumericDataError, MissingDataError

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

class ElasticNetSelector():
    def __init__(self, alpha_start : float = -2, alpha_finish : float = 0, beta_start : float = -2, beta_finish : float = 0, n_alphas : int = 100, extend_step : int = 20,power : int = 2) -> None:
        '''
            Setting the algorithm
        :param alpha_start: float, by default = -2
            The starting point in the greedy search of coefficients for L1 regularization
        :param alpha_finish: float, by default = 1
            The finish point in the greedy search of coefficients for L1 regularization
        :param beta_start: float, by default = -2
            The starting point in the greedy search of coefficients for L2 regularization
        :param beta_finish: float, by default = 1
            The finish point in the greedy search of coefficients for L2 regularization
        :param n_alphas: integer, by default = 100
            The number of points in greedy search
        :param extend_step: integer, by default = 20
            The quantity with which the :param alpha_start and :param alpha_finish will be updated
        :param power: integer, by default = 2
            Used to set a threshold in finding the best coefficients
        '''
        self.n_alphas = n_alphas
        self.alpha_start = alpha_start
        self.alpha_finish = alpha_finish
        self.beta_start = beta_start
        self.beta_finish = beta_finish
        self.extend_step = extend_step
        self.power = power

    def select(self,dataframe : pd.DataFrame, target : str, cv : int = 5) -> list:
        '''
            Selecting the most important columns
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: string
             The column name of the value that we what to predict
        :param cv: integer, by default = 5
            Determines the cross-validation splitting strategy
        :return: list
            The list of columns selected by algorithm
        '''
        # Checking if in the passed data are any NaN values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values!"
            )

        # Checking if in the passed data are non-numeric values.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Checking if the passed target column name is present in the passed data frame.
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist!"
            )

        #Splitting data
        self.dataframe = dataframe
        self.target = target
        self.X_columns = [col for col in self.dataframe.columns if col != self.target]
        X = self.dataframe[self.X_columns].values
        y = self.dataframe[self.target].values
        while True:
            #Setting alpha and beta values for greedy search
            self.alphas = np.logspace(self.alpha_start, self.alpha_finish, self.n_alphas, endpoint=True)
            self.betas = np.logspace(self.beta_start, self.beta_finish, self.n_alphas, endpoint=True)

            #Calculatina L1_ratio and alphas + reverse betas sum
            self.betas = self.betas[::-1]
            self.L1_ratio = self.alphas/(self.alphas+self.betas)
            self.alphas = self.alphas + self.betas

            #Running ElasticNetCV
            self.enet_cv = ElasticNetCV(cv=cv, alphas=self.alphas, l1_ratio=self.L1_ratio, random_state=0, tol=0.0001)
            self.enet_cv.fit(X, y)

            #Enlarge the interval if the optimal value is on one of the endpoints and repeat cycle
            if self.enet_cv.alpha_ == self.alphas[0]:
                self.alpha_start -= self.extend_step
                self.beta_start -= self.extend_step
            elif self.enet_cv.alpha_ == self.alphas[-1]:
                self.alpha_finish += self.extend_step
                self.beta_finish += self.extend_step
            else:
                #Selecting cols witch are more useful
                self.choosed_cols = [self.X_columns[i] for i in range(len(self.enet_cv.coef_)) if abs(self.enet_cv.coef_[i])>10 **(-self.power)]
                return self.choosed_cols
    
    
    def plot_process(self, eps : float = 5e-3, title :str = "ElasticNet coef Plot", save : bool = False, file_path : str = None, regularization_plot : str = 'L1'):
        '''
            Ploting the process of finding the best features
        :param eps: float, by default = 5e-3
            Length of the path
        :param title string, by default = "ElasticNet coef Plot"
            The title of the plot
        :param save boolean, by default = False
            If the this parameter is set to False that the model will not save the model
            If it is set to True the plot will be saved using :param file_path
        :param file_path: string, by default = None
            The file path where the plot will be saved
            If the :param save is set to False the it is not used
        :param file_path: string, by default = "L1"
            Two possible values: "L1" and "L2"
            For "L1" shows the plot for changes in L1 alphas
            For "L2" shows the plot for changes in L2 betas                    
        :return:
            Plots the process of the algorithm
        '''
        X = self.dataframe[self.X_columns].values
        y = self.dataframe[self.target].values

        #Plotting for L1 regularization
        if regularization_plot == 'L1':
            alpha = self.enet_cv.alpha_ * self.enet_cv.l1_ratio_

            #Getting the nearest values around the best value of alpha
            alphas = np.linspace(alpha-0.1, alpha+0.1, self.n_alphas, endpoint=True)
            alphas_enet, coefs_enet,_ = enet_path(X, y, eps, fit_intercept=False, alphas=alphas) 
            neg_log_alphas_enet = alphas_enet
            max_coef = coefs_enet[0][0]
            min_coef = coefs_enet[0][0]

            #Plotting the importance for every feature of the data
            for i in range(len(coefs_enet)):
                line_style = lambda col : '-' if col in self.choosed_cols else '--'
                plt.plot(neg_log_alphas_enet, coefs_enet[i], line_style(self.X_columns[i]), label=self.X_columns[i])
                if max(coefs_enet[i]) > max_coef:
                    max_coef = max(coefs_enet[i])
                if min(coefs_enet[i]) < min_coef:
                    min_coef = min(coefs_enet[i])
            plt.vlines(alpha, min_coef, max_coef, linestyles='dashed')
            plt.xlabel('-Log(alpha)')

        #Plotting for L1 regularization
        elif regularization_plot == "L2":
            beta =  self.enet_cv.alpha_ -  self.enet_cv.alpha_ * self.enet_cv.l1_ratio_

            #Getting the nearest values around the best value of beta
            betas = np.linspace(beta-0.1, beta+0.1, self.n_alphas, endpoint=True)
            betas_enet, coefs_enet, _ = enet_path(X, y, eps, fit_intercept=False, alphas=betas)
            neg_log_betas_enet = betas_enet
            max_coef = coefs_enet[0][0]
            min_coef = coefs_enet[0][0]

            #Plotting the importance for every feature of the data
            for i in range(len(coefs_enet)):
                line_style = lambda col : '-' if col in self.choosed_cols else '--'
                plt.plot(neg_log_betas_enet, coefs_enet[i], line_style(self.X_columns[i]), label=self.X_columns[i])
                if max(coefs_enet[i]) > max_coef:
                    max_coef = max(coefs_enet[i])
                if min(coefs_enet[i]) < min_coef:
                    min_coef = min(coefs_enet[i])
            plt.vlines(beta, min_coef, max_coef, linestyles='dashed')
            plt.xlabel('-Log(beta)')
        plt.ylabel('coefficients')
        plt.title(title)
        plt.axis('tight')
        plt.legend()
        if save:
            plt.savefig(file_path)
        plt.show()


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
import numpy as np
import matplotlib.pyplot as plt
from sklearn.linear_model import lasso_path, LassoCV
import pandas as pd
from .errors import MissingDataError, NonNumericDataError, NoSuchColumnError
def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn
class LassoSelector:
    def __init__(self, alpha_start : float = 0, alpha_finish : float = 2, n_alphas : int = 300, extend_step : float = 20, power : int = 2) -> None:
        '''
            Setting the algorithm
        :param alpha_start: float
            The starting point in the greedy search of coefficients
        :param alpha_finish: float
            The finish point in the greedy search of coefficients
        :param n_alphas: integer
            The number of points in greedy search
        :param extend_step: integer
            The quantity with which the :param alpha_start and :param alpha_finish will be updated
        :param power: integer
            Used to set a threshold in finding the best coefficients
        '''
        self.n_alphas = n_alphas
        self.alpha_start = alpha_start
        self.alpha_finish = alpha_finish
        self.extend_step = extend_step
        self.power = power

    def select(self, dataframe : 'pd.DataFrame', target : str, cv : int = 5) -> list:
        '''
            Selecting the most important columns
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: string
             The column name of the value that we what to predict
        :param cv: integer
            Determines the cross-validation splitting strategy
        :return: list
            The list of columns selected by algorithm
        '''
        # Checking if in the passed data are any NaN values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values!"
            )

        # Checking if in the passed data are non-numeric values.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Checking if the passed target column name is present in the passed data frame.
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist!"
            )
        self.dataframe = dataframe
        self.target = target
        self.X_columns = [col for col in self.dataframe.columns if col != self.target]

        # Converting the data frame into numpy arrays
        X = self.dataframe[self.X_columns].values
        y = self.dataframe[self.target].values

        # Searching for the optimal value of alpha
        while True:
            # Generating the log search space for alpha
            self.alphas = np.logspace(self.alpha_start, self.alpha_finish, self.n_alphas, endpoint=True)

            # Creating and fitting up the lasso CV
            self.lasso_cv = LassoCV(cv=cv, alphas=self.alphas, random_state=0, tol=0.01)
            self.lasso_cv.fit(X, y)

            # Deciding if we need to expand the search space of not
            if self.lasso_cv.alpha_ == self.alphas[0]:
                self.alpha_start -= self.extend_step
            elif self.lasso_cv.alpha_ == self.alphas[-1]:
                self.alpha_finish += self.extend_step
            else:
                # Selecting the columns with the coefficients bigger then 10e-power
                self.choosed_cols = [self.X_columns[i] for i in range(len(self.lasso_cv.coef_)) if abs(self.lasso_cv.coef_[i])>10 **(-self.power)]
                return self.choosed_cols
    def plot_process(self, eps : float = 5e-3, title : str = "Lasso coef Plot", save : bool =False, file_path : str = None) -> None:
        '''
            Ploting the process of finding the best features
        :param eps: float
            Length of the path
        :param title string
            The title of the plot
        :param save boolean, default = False
            If the this parameter is set to False that the model will not save the model
            If it is set to True the plot will be saved using :param file_path
        :param file_path: string, default = None
            The file path where the plot will be saved
            If the :param save is set to False the it is not used
        :return:
            Plots the process of the algorithm
        '''
        # Converting the data frame into numpy arrays
        X = self.dataframe[self.X_columns].values
        y = self.dataframe[self.target].values

        # Generating the line space for the X axis
        alphas = np.linspace(self.lasso_cv.alpha_-0.1, self.lasso_cv.alpha_+0.1, self.n_alphas, endpoint=True)

        # Generating the alphas and coefficients
        alphas_lasso, coefs_lasso, _ = lasso_path(X, y, eps, fit_intercept=False, alphas=alphas)
        neg_log_alphas_lasso = alphas_lasso
        max_coef = coefs_lasso[0][0]
        min_coef = coefs_lasso[0][0]

        # Plotting the lasso coefficients
        for i in range(len(coefs_lasso)):
            line_style = lambda col : '-' if col in self.choosed_cols else '--'
            plt.plot(neg_log_alphas_lasso, coefs_lasso[i], line_style(self.X_columns[i]), label=self.X_columns[i])
            if max(coefs_lasso[i]) > max_coef:
                max_coef = max(coefs_lasso[i])
            if min(coefs_lasso[i]) < min_coef:
                min_coef = min(coefs_lasso[i])
        plt.vlines(self.lasso_cv.alpha_, min_coef, max_coef, linestyles='dashed')
        plt.xlabel('-Log(alpha)')
        plt.ylabel('coefficients')
        plt.title(title)
        plt.axis('tight')
        plt.legend()
        if save:
            plt.savefig(file_path)
        plt.show()


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
import random
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split
import pandas as pd
from .errors import NoSuchColumnError, NonNumericDataError, MissingDataError

class GeneticAlgorithmSelector:
    def __init__(self, nb_children : int = 4, nb_generation : int = 200, scoring_metric = accuracy_score, max : bool = True) -> None:
        '''
            Setting up the model
        :param nb_children: integer, default = 4
            The number of children created at every generation
        :param nb_generation: integer, default = 200
            The number of generations created to find the best feature combination
        :param scoring_metric: function, default = sklearn.metrics.accuracy_score
            The sklearn like scoring metric
        :param max: boolean, default = True
            Used to define whatever we need the biggest of the smallest values computed by self.scoring_metric
        '''
        self.nb_children = nb_children
        self.nb_generation = nb_generation
        self.scoring_metric = scoring_metric
        self.max = max
    def inverse(self, value : int) -> int :
        '''
            This function inverses the value coming to function
        :param value: integer, o or 1
            The value that should be inverted
        :return: integer
            Return the inverse of :param value (if value = 0 ==> 1, else 0)
        '''
        if value == 0:
            return 1
        else:
            return 0
    def cross_over(self, population : list) -> list:
        '''
            This function apply the crossing-over process on 2 arrys-like lists
        :param population: 2-d array like list
            The population with 2 array that will be uses to create new individuals in the population
        :return: 2-d array like list
            Return the population with parents and their children after crossing-over
        '''
        # Creating an empty list for the new generation
        new_generation = []

        # Adding nb_children children to the new generatin
        for i in range(self.nb_children//2):
            first = random.randrange(0, len(self.X_columns)-1)
            second = random.randrange(0, len(self.X_columns)-1)
            if first > second:
                first, second = second, first
            new_generation.append(population[0][0:first] + population[1][first: second] + population[0][second:])
            new_generation.append(population[1][0:first] + population[0][first: second] + population[1][second:])

        # Adding the new generation to the population
        for gene in new_generation:
            population.append(gene)
        return population
    def mutate(self, gene : list) -> list:
        '''
            This function generates a random mutation on a gene
        :param gene: 1-d array like list
            The list with zeros and ones that will be mutated
        :return: 1-drray like list
            The gene list after mutation
        '''
        # Choosing a random place to generate a mutation
        mutation_locus = random.randrange(0, len(self.X_columns)-1)

        # Generating a mutation
        gene[mutation_locus] = self.inverse(gene[mutation_locus])
        return gene
    def gene_to_cols(self, gene : list) -> list:
        '''
            This function convert the zeros and ones list to columns list
        :param gene: 1-d array like list
            The list with zeros and ones that must be transformed in a columns sequence
        :return: 1-d array like list
            The lust with columns that will go to the model
        '''
        cols = []
        for i in range(len(gene)):
            if gene[i] == 1:
                cols.append(self.X_columns[i])
        return cols
    def select(self, algo, dataframe : 'pd.DataFrame', target : str, test_size=0.2):
        '''
            This function selects the best columns
        :param algo: sklearn algorithm class
            An sklearn algorithm class
        :param dataframe: pandas DataFrame
            Data Frame on which the algorithm is applied
        :param target: string
            The column name of the value that we what to predict
        :param test_size: float, default = 0.2
            The percent of data set that will be used to find the trained algorithm accuracy
        :return: list
            The list of columns selected by algorithm
        '''
        # Checking if in the passed data are any NaN values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values!"
            )

        # Checking if in the passed data are non-numeric values.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Checking if the passed target column name is present in the passed data frame.
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist!"
            )

        # Getting the list with names of columns without the target one
        self.X_columns = [col for col in dataframe.columns if col != target]

        # Generating the empty population and a temporal one
        population = []
        temp = []
        for i in range(len(self.X_columns)):
            temp.append(random.choice([0, 1]))
        population.append(temp)
        temp = [self.inverse(x) for x in population[-1]]
        population.append(temp)

        # Creating new generations and testing in searching for individual with the best metric
        for gen in range(self.nb_generation):
            # Creating an empty list with the metrics
            acc = []

            # Generating the first cross-over
            population = self.cross_over(population)

            # Generating some mutations
            for i in range(2, len(population)):
                population[i] = self.mutate(population[i])

            # Measuring the accuracy of every individual in the population
            for gene in population:
                if set(gene) == set([0]):
                    continue
                X = dataframe[self.gene_to_cols(gene)].values
                y = dataframe[target].values
                X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=0)
                algo.fit(X_train, y_train)
                y_pred = algo.predict(X_test)
                acc.append(self.scoring_metric(y_test, y_pred))

            # Selecting the 2 best members of the population based on whatever the metric's property
            if self.max:
                res = sorted(range(len(acc)), key=lambda sub: acc[sub])[-2:]
            else:
                res = sorted(range(len(acc)), key=lambda sub: acc[sub])[:2]

            # Creating a new population with the 2 best members
            population = [population[res[0]], population[res[1]]]
        return self.gene_to_cols(population[0])


'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
import numpy as np
import pandas as pd
from math import sqrt
from .errors import NotBetweenZeroAndOneError, NonNumericDataError, NoSuchColumnError, MissingDataError
class PointBiserialCorrSelector:
    def __init__(self, min_corr : float = 0.5, max_corr : float = 0.8, last_level : int = 2) -> None:
        '''
            Setting up the algorithm
        :param min_corr: float, between 0 and 1, default = 0.5
            The minimal positive correlation value that must the feature have with target
        :param max_corr: float, between 0 and 1, default = 0.8
            The maximal positive correlation value that must the feature have with target
        :param erase_corr: boolean, default = False
            If set as False the selector doesn't erase features that are highly correlated between themselves
            If set as True the selector does erase features that are highly correlated between themselves
        '''
        if min_corr < 0 or max_corr > 1 or min_corr > 1 or max_corr < 0:
            raise NotBetweenZeroAndOneError("Min or Max Correlations are not seted between 0 and 1!")
        else:
            self.min_corr = min_corr
            self.max_corr = max_corr
            self.last_level = last_level

    def pointBiserialCorr(self, dataframe : 'pd.DataFrame', X_col : str, y_col : str) -> float:
        '''
            Calculating the Point-Biserial Correlation
        :param dataframe: pandasDataFrame
            DataFrame from which we extract needed columns to calculate the correlation
        :param X_col: string
            The name of the column in DataFrame that holds continous data
        :param y_col: sting
            The name of the column in DataFrame that holds the binary Data
        :return:
            The calculated Point Biserial Correlation between X_col and y_col
        '''
        # Getting unique values from target column
        y_unique = dataframe[y_col].unique()

        # Sub-Setting data set
        subset0 = dataframe[dataframe[y_col] == y_unique[0]][X_col]
        subset1 = dataframe[dataframe[y_col] == y_unique[1]][X_col]

        # Getting the standard deviation for the column for the whole data set
        std = np.std(dataframe[X_col])

        # Getting the length of the sets
        n = len(dataframe[y_col])
        n0 = len(subset0)
        n1 = len(subset1)

        # Calculating the means of the both data set
        Mean0 = subset0.mean()
        Mean1 = subset1.mean()

        # Calculating the Point-Biserial correlation
        return (Mean0 - Mean1) * sqrt((n0*n1)/n**2) / std

    def select(self, dataframe : 'pd.DataFrame', target : str) -> list:
        '''
            Selecting the most important columns
        :param dataframe: pandas DataFrame
            Data Frame on which the algorithm is applied
        :param target: string
            The column name of the value that we what to predict
        :return: list
            The list of features that are selected by the algorithm as the best one
        '''
        # Checking if in the passed data are any NaN values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values!"
            )

        # Checking if in the passed data are non-numeric values.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Checking if the passed target column name is present in the passed data frame.
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist!"
            )

        # Defining the empty correlation and correlation-level dictionary
        corr_dict = {}
        corr_levels = {}

        # Creating the bountries list
        bountries = np.linspace(0, 1, 11, endpoint=True)

        # Getting the numerical columns
        numeric_cols_names = dataframe.select_dtypes(include=np.number).columns.tolist()

        # Calculating the point-biserial correlation for every numerical column
        for numeric_col in numeric_cols_names:
            if numeric_col != target:
                corr_dict[numeric_col] = self.pointBiserialCorr(dataframe, numeric_col, target)

        # Placing the columns in correlation levels
        for i in range(len(bountries)-1):
            corr_levels[str(bountries[i])] = []
            for numeric_col in corr_dict:
                if abs(corr_dict[numeric_col]) > bountries[i] and abs(corr_dict[numeric_col]) < bountries[i+1]:
                    corr_levels[str(bountries[i])].append(numeric_col)
        corr_levels = {x:corr_levels[x] for x in corr_levels if corr_levels[x]!=[]}
        levels = np.array(list(corr_levels.keys())).astype('float')

        # Getting the last_level highest correlated levels
        highest_levels = np.argsort(levels)[::-1][:self.last_level]
        self.high_corralated = []
        for level in highest_levels:
            self.high_corralated.extend(corr_levels[list(corr_levels.keys())[level]])

        # Returning the selected columns
        return self.high_corralated


'''
Created with love by Sigmoid
@Author - Basoc Nicoleta-Nina - nicoleta.basoc28@gmail.com
'''
import pandas as pd
import numpy as np
import phik
from .errors import MissingDataError, NonNumericDataError, NoSuchColumnError, NotBetweenZeroAndOneError

class PhikSelector:

    def __init__(self, min_corr = 0.3, max_corr = 0.8, erase_corr = True) -> None:
        """
            Setting up the model
        :param min_corr: float, between 0 and 1, default = 0.5
            The minimal positive correlation value that must the feature have with y_column
        :param max_corr: float, between 0 and 1, default = 0.8
            The maximal positive correlation value that must the feature have with y_column
        :param erase_corr: boolean, default = False
            If set as False the selector doesn't erase features that are highly correlated between themselves
            If set as True the selector does erase features that are highly correlated between themselves
        """
        if min_corr < 0 or max_corr > 1 or min_corr > 1 or max_corr < 0:
            raise NotBetweenZeroAndOneError("Min or Max Correlations are not seted between 0 and 1!")
        else:
            self.min_corr = min_corr
            self.max_corr = max_corr
        if str(type(erase_corr)) != "<class 'bool'>":
            raise ValueError("Parameter erase_col isn't set as a boolean value!")
        else:
            self.erase_corr = erase_corr
    
    def index_to_cols(self, index_list : int) -> list:
        '''
            Converting the indexes list into names of features that are included in model
        :param index_list: list
            A list of Indexes that should be converted into a list of feature names
        :return: list
            A list with feature names
        '''
        return [self.X_columns[i] for i in index_list]
        
    def select(self, dataframe : pd.DataFrame, target : str) -> list:
        """
            Selecting the most important columns
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: str
             The column name of the value that we what to predict
        :return: list
            The list of features that our algorithm selects as the best ones
        """
        
        # Verifying if data frame has NaN-values.
        if dataframe.isna().values.any():
            raise MissingDataError("The passed data frame contains missing values")
        
        # If the given target value doesn't match any column in the DataFrame raise an error
        if target not in dataframe:
            raise NoSuchColumnError(f'No such column {target}')
        
        # If a column has string values raise an error.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values')
                
        # Getting the list with names of columns without the target one.
        self.X_columns = [col for col in dataframe.columns if col != target]
        
        # Getting the feature columns.
        X = dataframe.drop([target], axis=1).columns
        
        # Creating an empty list for correlated columns.
        correlated_indexes = []
        
        # Initialising the phik correlation on the dataframe   . 
        dataframe.corr()
        
        # Getting the target-column index.
        target_index = list(dataframe.columns).index(target)
        # Getting the correlation matrix.
        phik_corr_matrix= dataframe.phik_matrix(interval_cols=X)
        
        # Generating the correlation matrix
        self.corr_table = dataframe.phik_matrix(interval_cols=X)
        corr_matrix = self.corr_table.values
    
        # Searching for the columns correlated with the target-column.
        for i in range(len(corr_matrix[target_index])):
            if abs(corr_matrix[target_index][i]) > self.min_corr and abs(corr_matrix[target_index][i]) < self.max_corr:
                correlated_indexes.append(i)
                
        # Creating a list with the names of columns correlated with the y-column.
        self.correlated_cols = self.index_to_cols(correlated_indexes)
        
        # If we chose to erase a column from the correlated pairs we erase one of them.
        if self.erase_corr:
            cols_to_remove = []
            for i in correlated_indexes:
                for j in correlated_indexes:
                    if abs(corr_matrix[i][j]) > self.max_corr and abs(corr_matrix[i][j]) < self.max_corr:
                        cols_to_remove.append(self.X_columns[i])
            cols_to_remove = set(cols_to_remove)
            # Removing the chosen columns.
            for col in cols_to_remove:
                self.correlated_cols.remove(col)
                
        return self.correlated_cols



'''
Created with love by Sigmoid.
@Author - Spînu Lavinia - laviniaspinu01@gmail.com
'''
# Import all needed libraries.
import pandas as pd
from sklearn.decomposition import FastICA
from .errors import MissingDataError, NonNumericDataError, NoSuchColumnError, NotBetweenZeroAndOneError, DifferentColumnsError
def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn


class ICAReducer():

    def __init__(self, min_corr : float = 0.5, max_corr : float = 0.8, correlation_type : str = "pearson"):
        '''
            Setting up the algorithm
        :param min_corr: float, between 0 and 1, default = 0.5
            The minimal positive correlation value that must the feature have with target
        :param max_corr: float, between 0 and 1, default = 0.8
            The maximal positive correlation value that must the feature have with target
        :param correlation_type: str, default = "pearson"
        '''
        if min_corr < 0 or min_corr > 1:
            raise NotBetweenZeroAndOneError("min_corr isn't between 0 and 1")
        else:
            self.min_corr = min_corr
        if max_corr < 0 or max_corr > 1:
            raise NotBetweenZeroAndOneError("max_corr isn't between 0 and 1")
        else:
            self.max_corr = max_corr
        if correlation_type in ['pearson', 'kendall', 'spearman']:
            self.correlation_type = correlation_type
        else:
            raise NoSuchMethodError(
                f"kydavra doesn't sustain such method as {correlation_type}/nTry 'pearson', 'kendall' or 'spearman'")

    def index_to_cols(self, index_list):
        '''
            Converting the list of indexes in a list of features that will be picked by the model
        :param index_list: list
            A list of Indexes that should be converted into a list of feature names
        :return: list
            A list with feature names
        '''
        return [self.X_columns[i] for i in index_list]


    def reduce(self,dataframe : pd.DataFrame, target : str):
        '''
            Reducing the dimensionality of the data
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: string
             The column name of the value that we have to predict
        '''
        # Checking for missing values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values."
            )

        # if a column has string values raise an error
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # if the given target doesn't match any column in the DataFrame raise an error
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist."
            )

        # Setting variables
        self.data_frame = dataframe.copy()
        self.dataframe = dataframe.copy()
        self.X_columns = [col for col in self.dataframe.columns if col != target]
        correlated_indexes = []

        # Getting index of label column
        target_index = list(self.dataframe.columns).index(target)

        # Creating correlation table
        self.corr_table = self.dataframe.corr()
        corr_matrix = self.corr_table.values

        # Choosing most important columns
        for i in range(len(corr_matrix[target_index])):
            if abs(corr_matrix[target_index][i]) > self.min_corr and abs(corr_matrix[target_index][i]) < self.max_corr:
                correlated_indexes.append(i)
        self.correlated_cols = self.index_to_cols(correlated_indexes)

        # Creating correlation table for selected columns
        correlated_features_df = self.dataframe[self.correlated_cols]
        correlated_features_matrix = correlated_features_df.corr()
        self.correlated_features = []

        # Selecting most important correlations between features and creting lists of correlations
        for i in range(len(correlated_features_matrix)):
            cols = []
            for j in range(len(correlated_features_matrix)):
                # I also select the col on diagonal because it is also a part of correlation and should be in my array
                if abs(correlated_features_matrix[self.correlated_cols[i]][self.correlated_cols[j]]) > self.min_corr and abs(correlated_features_matrix[self.correlated_cols[i]][self.correlated_cols[j]]) < self.max_corr :
                    cols.append(self.correlated_cols[j])
            if len(cols) != 0 :
                cols.append(self.correlated_cols[i])
                cols.sort()
                self.correlated_features.append(cols)

        # Cleaning selected columns and ordering them by length
        self.correlated_features = [list(item) for item in set(tuple(row) for row in self.correlated_features)]
        self.correlated_features = sorted(self.correlated_features, key=len, reverse=True)
        self.dictionary = {}
        self.correlated_columns = []

        # Creating main dictionary with ICA filters for different correlations
        for i in range(len(self.correlated_features)):
            flag = False
            key = tuple(self.correlated_features[i])
            for j in range(len(self.correlated_features[i])):
                if self.correlated_features[i][j] not in self.dataframe.columns:
                    flag = True
            if flag:
                continue

            # Creating algorithm for every correlation
            self.correlated_columns.append(self.correlated_features[i]) # addding really changed columns
            correlated_features_ICA_df = self.dataframe[self.correlated_features[i]].values
            ica = FastICA(n_components = 1)
            ica.fit(correlated_features_ICA_df)
            X_ica = ica.transform(correlated_features_ICA_df)

            # Creating new column and adding it to dataframe
            column_name = "_".join(key)
            self.dataframe[column_name] = X_ica
            self.dictionary[column_name] = ica

            # Deleting correlated columns
            for j in range(len(self.correlated_features[i])):
                self.dataframe = self.dataframe.drop(self.correlated_features[i][j],1)
        return self.dataframe

    def apply(self, dataframe : pd.DataFrame):
        '''
            Reducing the dimensionality of the data
            based on an already existed filter
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: string
             The column name of the value that we have to predict
        '''
        # Checking for missing values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values."
            )

        # if a column has string values raise an error
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Splitting the data.
        self.dataframe = dataframe.copy()

        # Checking if the names of the columns are different for the both dataframes.
        if list(self.data_frame.columns) != list(self.dataframe.columns):
            raise DifferentColumnsError('The passed dataframe has different columns from the one passed to filter function.')
        for i in range(len(self.correlated_columns)):
            correlated_features_ICA_df = self.dataframe[self.correlated_columns[i]].values
            column_name = "_".join(self.correlated_columns[i])
            ica = self.dictionary[column_name]
            X_ica = ica.transform(correlated_features_ICA_df)
            self.dataframe[column_name] = X_ica
            for j in range(len(self.correlated_features[i])):
                self.dataframe = self.dataframe.drop(self.correlated_columns[i][j],1)
        return self.dataframe



'''
Created with love by Sigmoid
@Author - iurie.cius.personal@gmail.com
'''

# Importing all needed libraries.
import numpy as np
import pandas as pd
from .errors import MissingDataError, NonNumericDataError, NoSuchColumnError

class FisherSelector:
    def __init__(self, n_features: int = 5) -> None:
        """
            Setting up the algorithm
        :param n_features: int, default = 5
            The number of top features (according to the relieff score) to
            retain after feature selection is applied.
        """
        self.n_features = n_features

    def select(self, dataframe: 'pd.DataFrame', target: str) -> list:
        """
            Selecting the most important columns.
        :param dataframe: pandas DataFrame
            Data Frame on which the algorithm is applied.
        :param target: str
            The column name of the value that we what to predict.
        :return: list
            The list of features that are selected by the algorithm as the best one.
        """

        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values.")

        # if a column has string values raise an error
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError(
                    'The given DataFrame contains non-numeric values !')

        if target not in dataframe:
            raise NoSuchColumnError("The passed target column doesn't exist.")

        # Converting the data frame in numpy arrays.
        X = dataframe.drop([target], axis=1).values
        y = dataframe[target].values

        # Get the dimensionality of the DataFrame.
        n_samples, n_features = X.shape

        # Initialize the scores to 0.
        score = np.zeros(n_features)

        # Get the number of classes.
        classes = np.unique(y).tolist()

        # Initialize all the variables to 0.
        n_k, mean_k, mean_dataframe, std_dataframe = 0, 0, 0, 0

        # Iterating through every feature in the dataset.
        for i in range(n_features):

            # Iterating through every class in the dataset.
            for c in classes:
                # Compute the size of the k-th class.
                n_k = dataframe[target].value_counts().get(c)

                # Compute the mean of the k-th class.
                mean_k = dataframe.groupby(target).mean()[
                    dataframe.columns[i]].get(c)

                # Compute the mean of the whole dataset.
                mean_dataframe = dataframe[dataframe.columns[i]].mean()

                # Compute the standard deviation of the whole dataset.
                std_dataframe = dataframe[dataframe.columns[i]].std()

                # Update the score.
                score[i] += (n_k * (mean_k - std_dataframe)**2)

            # Update the score
            score[i] /= mean_dataframe**2

        # Sorting the indexes based on the score.
        res = np.argsort(score, 0)
        idx = res[::-1][:self.n_features]

        # Return the top n features.
        return list(dataframe.iloc[:, idx].columns)



'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
from .errors import NotBetweenZeroAndOneError, MissingDataError, NonNumericDataError, NoSuchColumnError
import pandas as pd
class KendallCorrelationSelector:
    def __init__(self, min_corr : float = 0.5, max_corr : float = 0.8, erase_corr : bool = False) -> None:
        '''
            Setting up the model
        :param min_corr: float, between 0 and 1, default = 0.5
            The minimal positive correlation value that must the feature have with target
        :param max_corr: float, between 0 and 1, default = 0.8
            The maximal positive correlation value that must the feature have with target
        :param erase_corr: boolean, default = False
            If set as False the selector doesn't erase features that are highly correlated between themselves
            If set as True the selector does erase features that are highly correlated between themselves
        '''
        if min_corr < 0 or max_corr > 1 or min_corr > 1 or max_corr < 0:
            raise NotBetweenZeroAndOneError("Min or Max Correlations are not seted between 0 and 1!")
        else:
            self.min_corr = min_corr
            self.max_corr = max_corr
        if str(type(erase_corr)) != "<class 'bool'>":
            raise ValueError("Parameter erase_col isn't set as a boolean value!")
        else:
            self.erase_corr = erase_corr
    def index_to_cols(self, index_list : int) -> list:
        '''
            Converting the indexes list into names of features that are included in model
        :param index_list: list
            A list of Indexes that should be converted into a list of feature names
        :return: list
            A list with feature names
        '''
        return [self.X_columns[i] for i in index_list]
    def select(self, dataframe : 'pd.DataFrame', target : str) -> list:
        '''
            Selecting the most important columns
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: string
             The column name of the value that we what to predict
        :return: list
            The list of features that are selected by the algorithm as the best one
        '''
        # Checking if in the passed data are any NaN values.
        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values!"
            )

        # Checking if in the passed data are non-numeric values.
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # Checking if the passed target column name is present in the passed data frame.
        if target not in dataframe:
            raise NoSuchColumnError(
                "The passed target column doesn't exist!"
            )

        # Getting the list with names of columns without the target one
        self.X_columns = [col for col in dataframe.columns if col != target]

        # Creating an empty list for correlated columns
        correlated_indexes = []

        # Getting the y-column index
        target_index = list(dataframe.columns).index(target)

        # Generating the correlation matrix
        self.corr_table = dataframe.corr(method='kendall')
        corr_matrix = self.corr_table.values

        # Searching for the columns correlated with the y-column
        for i in range(len(corr_matrix[target_index])):
            if abs(corr_matrix[target_index][i]) > self.min_corr and abs(corr_matrix[target_index][i]) < self.max_corr:
                correlated_indexes.append(i)

        # Creating a list with the names of columns correlated with the y-column
        self.correlated_cols = self.index_to_cols(correlated_indexes)

        # If we chose to erase a column from the correlated pairs we erase one of them
        if self.erase_corr:
            cols_to_remove = []
            for i in correlated_indexes:
                for j in correlated_indexes:
                    if abs(corr_matrix[i][j]) > self.min_corr or abs(corr_matrix[i][j]) < self.max_corr:
                        cols_to_remove.append(self.X_columns[i])
            cols_to_remove = set(cols_to_remove)

            # Removing the chose columns
            for col in cols_to_remove:
                self.correlated_cols.remove(col)
        return self.correlated_cols



"""
Created with love by Sigmoid
@Author - Smocvin Denis - denissmocvin@gmail.com
"""

import numpy as np
import pandas as pd
import math
from .errors import MissingDataError, NonNumericDataError, NoSuchColumnError

class BregmanDivergenceSelector:
    def __init__(self, min_divergence: int = 0, max_divergence: int = 10):
        """
        Initialize the algorithm.
        :param min_divergence:
            The minimum divergence that the columns should have
        :param max_divergence:
            The maximum divergence that the columns should have
        """

        self.min_divergence = min_divergence
        self.max_divergence = max_divergence

    def __calculate_divergence(self, q: 'pd.Series', p: 'pd.Series') -> float:
        """
        Calculating the Bregman divergence between a feature array and a target array.
        :param q: pandas Series
            Feature array
        :param p: pandas Series
            Target array
        :return: float
            The calculated Bregman divergence
        """

        d = 0

        # calculating the divergence
        for i in range(len(p)):
            d += math.sqrt(p[i] ** 2 + q[i] ** 2)

        # returning the normalized value of the divergence
        return d / len(p)

    def select(self, dataframe: 'pd.Dataframe', target: str) -> list:
        """
        Selecting columns according to their divergence.
        :param dataframe: pandas DataFrame
            The DataFrame on which the selection is applied
        :param target: string
            The name of the target column
        :return: list
            The list of columns that have a divergence relative to the target column between min_divergence and max_divergence
        """

        self.dataframe = dataframe
        self.X_columns = [col for col in self.dataframe.columns if col != target]

        # if the given y_column doesn't match any column in the DataFrame raise an error
        if target not in self.dataframe:
            raise NoSuchColumnError(f'No such column {target}')

        # if a column has string values raise an error
        for col in self.dataframe:
            if not pd.api.types.is_numeric_dtype(self.dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # raise an error if there are NaN values in the DataFrame
        if self.dataframe.isna().values.any():
            raise MissingDataError('The given DataFrame contains NaN values !')

        # defining a list to store the names of selected columns
        selected_cols = []

        for col in self.X_columns:
            # calculating the Bregman divergence of the column
            d = self.__calculate_divergence(self.dataframe[col], self.dataframe[target])

            # selecting the column if its divergence is between min_divergence and max_divergence
            if self.min_divergence < d < self.max_divergence:
                selected_cols.append(col)

        return selected_cols